/*
 * NeuralRackvst.cpp
 *
 * Binds the generic VST2 wrapper to the NeuralRack PluginClient.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 * 
 * Copyright (C) 2026 brummer <brummer@web.de>
 */

#define VST2IPLUG

#include "PluginClient.cc"
#include "VST2Wrapper.cpp"
